/* Copyright (C) 2010 Broadcom Corporation.
 * Portions Copyright (C) VMware, Inc. 2007-2010. All Rights Reserved.
 */

#define TG3_NETQ_WAIT_EVENT_TIMEOUT	msecs_to_jiffies(100)
#if 0
#define TG3_VRQ_MAX_NUM_TX_QS(tp)	(min_t(u16, (tp)->irq_cnt - 1, 15))
#define TG3_VRQ_MAX_NUM_RX_QS(tp)	((u16) tp->irq_cnt - 1)
#else
#define TG3_VRQ_MAX_NUM_TX_QS(tp)	0
#define TG3_VRQ_MAX_NUM_RX_QS(tp)	0
#endif

static void tg3_vmware_timer(struct tg3 *tp)
{
	/*
	 * If the driver says this NIC is in promiscuous mode, the hardware
	 * better agree with it. See PR 88299 for more details, but there
	 * are cases (eg. bcm5704 rev10) that the set promiscuous bit is
	 * lost for no reason and this seems to be the least intrusive
	 * workaround for that issue (likely a hardware bug).
	 *
	 * This is happening on other cards as well (BCM5700 rev04) PR 97750
	 * Also, on some cards (5703) we have seen other bits getting messed up
	 */
	if (netif_carrier_ok(tp->dev)) {
		u32 rx_mode = tr32(MAC_RX_MODE);
		if (!(rx_mode & RX_MODE_PROMISC) && (rx_mode != tp->rx_mode)) {
			/*
			 * We love to warn the users every time there is such a
			 * register reset, but we do not want to do it forever
			 * given some NICs are really that bad (PR 105488).
			 */
			if (tp->vmware.rx_mode_reset_counter < 200) {
				netdev_info(tp->dev, "%s: rx_mode "
					    "0x%x(%s)=>0x%x tg3_flags 0x%x "
					    "tg3_flags2 0x%x\n",
					    tp->dev->name, rx_mode,
					    rx_mode & RX_MODE_PROMISC ? "on" : "off",
					    tp->rx_mode, tp->tg3_flags,
					    tp->tg3_flags2);
				tp->vmware.rx_mode_reset_counter++;
			}
			tw32_f(MAC_RX_MODE, tp->rx_mode);
		}
	}

	/*
	 * we've seen ssome 5703's (rev 10) which don't seen to
	 * generate an interrupt on link-up state changes. bug 89197.
	 */
	if (!netif_carrier_ok(tp->dev) &&
	    !(tp->tg3_flags &
	     (TG3_FLAG_USE_LINKCHG_REG | TG3_FLAG_POLL_SERDES))) {
		struct tg3_hw_status *sblk = tp->napi[0].hw_status;
		if (sblk->status & SD_STATUS_LINK_CHG) {
			sblk->status = SD_STATUS_UPDATED |
			       (sblk->status & ~SD_STATUS_LINK_CHG);
			tg3_setup_phy(tp, 0);
		}
	}
}

#if !defined(TG3_VMWARE_BMAPILNX_DISABLE)
static int
tg3_vmware_ioctl_cim(struct net_device *dev, struct ifreq *ifr)
{
	struct tg3 *tp = netdev_priv(dev);
	void __user *useraddr = ifr->ifr_data;
	struct brcm_vmware_ioctl_req req;
	int rc = 0;
	u32 val;

	if (copy_from_user(&req, useraddr, sizeof(req))) {
		netdev_err(dev, "%s: could not copy "
				"from user tg3_ioctl_req\n", __func__);
		return -EFAULT;
	}

	switch(req.cmd) {
	case BRCM_VMWARE_CIM_CMD_ENABLE_NIC:
		netdev_info(dev, "%s: enable NIC\n", __func__);

		rc = tg3_open(tp->dev);
		break;
	case BRCM_VMWARE_CIM_CMD_DISABLE_NIC:
		netdev_info(dev, "%s: disable NIC\n", __func__);

		rc = tg3_close(tp->dev);
		break;
	case BRCM_VMWARE_CIM_CMD_REG_READ: {
		struct brcm_vmware_ioctl_reg_read_req *rd_req;

		rd_req = &req.cmd_req.reg_read_req;
		if (0x7ffc < rd_req->reg_offset) {
			netdev_err(dev, "%s: %s: "
					"out of range: req reg: 0x%x\n",
				   __func__, "reg read", rd_req->reg_offset);
			rc = -EINVAL;
			break;
		}

		if (rd_req->reg_offset & 0x3) {
			netdev_err(dev, "%s: %s: "
				   "offset not dword aligned: req reg: 0x%x\n",
				   __func__, "reg read", rd_req->reg_offset);
			rc = -EINVAL;
			break;
		}

		switch (rd_req->reg_access_type) {
		case BRCM_VMWARE_REG_ACCESS_DIRECT:
			val = tr32(rd_req->reg_offset);
			break;
		case BRCM_VMWARE_REG_ACCESS_PCI_CFG:
			pci_read_config_dword(tp->pdev,
					      rd_req->reg_offset, &val);
			break;
		default:
			netdev_err(dev, "%s: %s: "
				  "invalid access method: access type: 0x%x\n",
				   __func__, "reg read",
				   rd_req->reg_access_type);
			rc = -EINVAL;
			break;
		}

		req.cmd_req.reg_read_req.reg_value = val;
		netdev_info(dev, "%s: %s: reg: 0x%x value:0x%x",
			    __func__, "reg read", rd_req->reg_offset,
			    rd_req->reg_value);

		break;
	} case BRCM_VMWARE_CIM_CMD_REG_WRITE: {
		struct brcm_vmware_ioctl_reg_write_req *wr_req;

		wr_req = &req.cmd_req.reg_write_req;
		if (0x7ffc < wr_req->reg_offset) {
			netdev_err(dev, "%s: %s: "
					"out of range: req reg: 0x%x\n",
				   __func__, "reg write", wr_req->reg_offset);
			rc = -EINVAL;
			break;
		}

		if (wr_req->reg_offset & 0x3) {
			netdev_err(dev, "%s: %s: "
				   "offset not dword aligned: req reg: 0x%x\n",
				   __func__, "reg write", wr_req->reg_offset);
			rc = -EINVAL;
			break;
		}

		switch (wr_req->reg_access_type) {
		case BRCM_VMWARE_REG_ACCESS_DIRECT:
			tw32(wr_req->reg_offset, wr_req->reg_value);
			break;
		case BRCM_VMWARE_REG_ACCESS_PCI_CFG:
			pci_write_config_dword(tp->pdev, wr_req->reg_offset,
					       wr_req->reg_value);
			break;
		default:
			netdev_err(dev, "%s: %s: "
				  "invalid access method: access type: 0x%x\n",
				   __func__, "reg write",
				   wr_req->reg_access_type);
			rc = -EINVAL;
			break;
		}

		netdev_info(dev, "%s: %s: reg: 0x%x value:0x%x",
			    __func__, "reg write", wr_req->reg_offset,
			    wr_req->reg_value);

		break;
	} case BRCM_VMWARE_CIM_CMD_GET_NIC_PARAM:
		netdev_info(dev, "%s: get NIC param\n", __func__);

		req.cmd_req.get_nic_param_req.mtu = dev->mtu;
		memcpy(req.cmd_req.get_nic_param_req.current_mac_addr,
		       dev->dev_addr,
		       sizeof(req.cmd_req.get_nic_param_req.current_mac_addr));
		break;
	case BRCM_VMWARE_CIM_CMD_GET_NIC_STATUS:
		netdev_info(dev, "%s: get NIC status\n", __func__);

		req.cmd_req.get_nic_status_req.nic_status = netif_running(dev);
		break;
	default:
		netdev_err(dev, "%s: unknown req.cmd: 0x%x\n",
			   __func__, req.cmd);
		rc = -EINVAL;
	}

	if (rc == 0 &&
	    copy_to_user(useraddr, &req, sizeof(req))) {
		netdev_err(dev, "%s: couldn't copy to user tg3_ioctl_req\n",
			   __func__);
                return -EFAULT;
	}

	return rc;
}
#endif  /* TG3_VMWARE_BMAPILNX */

#ifdef TG3_VMWARE_NETQ_ENABLE
static void tg3_set_prod_bdinfo(struct tg3 *tp, u32 bdinfo_addr,
				dma_addr_t mapping, u32 maxlen_flags)
{
	tw32(bdinfo_addr + TG3_BDINFO_HOST_ADDR + TG3_64BIT_REG_HIGH,
	     ((u64) mapping >> 32));
	tw32(bdinfo_addr + TG3_BDINFO_HOST_ADDR + TG3_64BIT_REG_LOW,
	     ((u64) mapping & 0xffffffff));
	tw32(bdinfo_addr + TG3_BDINFO_MAXLEN_FLAGS, maxlen_flags);
	/* Leave the nic addr field alone */
}

static void tg3_disable_prod_rcbs(struct tg3 *tp)
{
	int i;
	u32 offset = RCVDBDI_JMB_BD_RING1;

	if (!(tp->tg3_flags3 & TG3_FLG3_IOV_CAPABLE))
		return;

	/* Disable all receive producer rings but the first. */
	for (i = 0; i < 16; i++) {
		/* Disable the jumbo ring */
		tg3_set_prod_bdinfo(tp, offset, 0, BDINFO_FLAGS_DISABLED);
		offset += TG3_BDINFO_SIZE;

		/* Disable the standard ring */
		tg3_set_prod_bdinfo(tp, offset, 0, BDINFO_FLAGS_DISABLED);
		offset += TG3_BDINFO_SIZE;
	}
}

static int
tg3_netq_get_netqueue_features(vmknetddi_queueop_get_features_args_t *args)
{
	args->features = VMKNETDDI_QUEUEOPS_FEATURE_NONE;
	/* args->features |= VMKNETDDI_QUEUEOPS_FEATURE_TXQUEUES; */
	return VMKNETDDI_QUEUEOPS_OK;
}

static int
tg3_netq_get_queue_count(vmknetddi_queueop_get_queue_count_args_t *args)
{
	struct tg3 *tp = netdev_priv(args->netdev);

	if (args->type == VMKNETDDI_QUEUEOPS_QUEUE_TYPE_RX) {
		args->count = TG3_VRQ_MAX_NUM_RX_QS(tp);

		netdev_info(tp->dev, "Using %d RX NetQ rings\n", args->count);

		return VMKNETDDI_QUEUEOPS_OK;
	} else if (args->type == VMKNETDDI_QUEUEOPS_QUEUE_TYPE_TX) {
		args->count = TG3_VRQ_MAX_NUM_TX_QS(tp);

		netdev_info(tp->dev, "Using %d TX NetQ rings\n", args->count);

		return VMKNETDDI_QUEUEOPS_OK;
	} else {
		netdev_err(tp->dev, "Counting queue: invalid queue type\n");

		return VMKNETDDI_QUEUEOPS_ERR;
	}
}

static int
tg3_netq_alloc_tx_queue(struct net_device *netdev,
			 vmknetddi_queueops_queueid_t *p_qid,
			 u16 *queue_mapping)
{
	struct tg3 *tp = netdev_priv(netdev);
#if 0
	int i;

	if (tp->vmware.netq.n_tx_queues_allocated >= TG3_VRQ_MAX_NUM_TX_QS(tp))
		return VMKNETDDI_QUEUEOPS_ERR;

	for (i = 1; i < TG3_VRQ_MAX_NUM_TX_QS(tp) + 1; i++) {
		struct tg3_napi *tnapi = &tp->napi[i];
		if (!tnapi->netq.tx_queue_allocated) {
			tnapi->netq.tx_queue_allocated = true;
			tp->vmware.netq.n_tx_queues_allocated++;
			*p_qid = VMKNETDDI_QUEUEOPS_MK_TX_QUEUEID(i);
			*queue_mapping = i;

			netdev_info(tp->dev, "TX NetQ allocated on %d\n", i);
			return VMKNETDDI_QUEUEOPS_OK;
		}
	}
#endif

	netdev_err(tp->dev, "No free tx queues found!\n");
	return VMKNETDDI_QUEUEOPS_ERR;
}

static int
tg3_netq_alloc_queue(vmknetddi_queueop_alloc_queue_args_t *args)
{
	struct net_device *netdev = args->netdev;
	struct tg3 *tp = netdev_priv(netdev);

	if (args->type == VMKNETDDI_QUEUEOPS_QUEUE_TYPE_TX) {
		return tg3_netq_alloc_tx_queue(args->netdev, &args->queueid,
					     &args->queue_mapping);
	} else {
		netdev_err(tp->dev, "Trying to alloc invalid queue type: %x\n",
			   args->type);
		return VMKNETDDI_QUEUEOPS_ERR;
	}
}

static int
tg3_netq_free_tx_queue(struct net_device *netdev,
		       vmknetddi_queueops_queueid_t qid)
{
	struct tg3 *tp = netdev_priv(netdev);
	struct tg3_napi *tnapi;
	int rc;

	u16 index = VMKNETDDI_QUEUEOPS_QUEUEID_VAL(qid);
	if (index == 0 || index > TG3_VRQ_MAX_NUM_TX_QS(tp))
		return VMKNETDDI_QUEUEOPS_ERR;

	tnapi = &tp->napi[index];

	if (!tnapi->netq.tx_queue_allocated)
		return VMKNETDDI_QUEUEOPS_ERR;

	tnapi->netq.tx_queue_allocated = false;
	tp->vmware.netq.n_tx_queues_allocated--;

	/* Wait for the TX to flush all packets */
	/* NOTE : Assumes no more packets will be submitted to this queue. */
	tnapi->netq.flags |= TG3_NETQ_FREE_TX_QUEUE_STATE;
	rc = wait_event_timeout(tp->vmware.netq.tx_wait,
				!(tnapi->netq.flags &
				  TG3_NETQ_FREE_TX_QUEUE_STATE),
				TG3_NETQ_WAIT_EVENT_TIMEOUT);
	tnapi->netq.flags &= ~TG3_NETQ_FREE_TX_QUEUE_STATE;

	if (tnapi->tx_cons != tnapi->tx_prod) {
		netdev_warn(tp->dev,
			    "Timeout submitting free NetQ TX Queue: %x\n",
			    index);
		return VMKNETDDI_QUEUEOPS_ERR;
	}

	netdev_info(tp->dev, "Free NetQ TX Queue: %x\n", index);

	return VMKNETDDI_QUEUEOPS_OK;
}

static int
tg3_netq_free_queue(vmknetddi_queueop_free_queue_args_t *args)
{
	struct tg3 *tp = netdev_priv(args->netdev);
	if (VMKNETDDI_QUEUEOPS_IS_TX_QUEUEID(args->queueid)) {
		return tg3_netq_free_tx_queue(args->netdev, args->queueid);
	} else {
		netdev_err(tp->dev, "free netq: invalid queue type\n");
		return VMKNETDDI_QUEUEOPS_ERR;
	}
}

static int
tg3_netq_get_queue_vector(vmknetddi_queueop_get_queue_vector_args_t *args)
{
	int qid;
	struct tg3 *tp = netdev_priv(args->netdev);
	qid = VMKNETDDI_QUEUEOPS_QUEUEID_VAL(args->queueid);
	if (qid > tp->irq_cnt - 1)
		return VMKNETDDI_QUEUEOPS_ERR;

	args->vector = tp->napi[qid].irq_vec;

	return VMKNETDDI_QUEUEOPS_OK;
}

static int
tg3_netq_get_default_queue(vmknetddi_queueop_get_default_queue_args_t *args)
{
	if (args->type == VMKNETDDI_QUEUEOPS_QUEUE_TYPE_TX) {
		args->queueid = VMKNETDDI_QUEUEOPS_MK_TX_QUEUEID(0);
		args->queue_mapping = 0;
		return VMKNETDDI_QUEUEOPS_OK;
	} else
		return VMKNETDDI_QUEUEOPS_ERR;
}

static int
tg3_netqueue_ops(vmknetddi_queueops_op_t op, void *args)
{
	struct tg3 *tp;

	if (op == VMKNETDDI_QUEUEOPS_OP_GET_VERSION)
		return vmknetddi_queueops_version(
			(vmknetddi_queueop_get_version_args_t *)args);

	tp = netdev_priv(((vmknetddi_queueop_get_features_args_t *)args)->netdev);

	switch (op) {
	case VMKNETDDI_QUEUEOPS_OP_GET_FEATURES:
		return tg3_netq_get_netqueue_features(
			(vmknetddi_queueop_get_features_args_t *)args);

	case VMKNETDDI_QUEUEOPS_OP_GET_QUEUE_COUNT:
		return tg3_netq_get_queue_count(
			(vmknetddi_queueop_get_queue_count_args_t *)args);

	case VMKNETDDI_QUEUEOPS_OP_ALLOC_QUEUE:
		return tg3_netq_alloc_queue(
			(vmknetddi_queueop_alloc_queue_args_t *)args);

	case VMKNETDDI_QUEUEOPS_OP_FREE_QUEUE:
		return tg3_netq_free_queue(
			(vmknetddi_queueop_free_queue_args_t *)args);

	case VMKNETDDI_QUEUEOPS_OP_GET_QUEUE_VECTOR:
		return tg3_netq_get_queue_vector(
			(vmknetddi_queueop_get_queue_vector_args_t *)args);

	case VMKNETDDI_QUEUEOPS_OP_GET_DEFAULT_QUEUE:
		return tg3_netq_get_default_queue(
			(vmknetddi_queueop_get_default_queue_args_t *)args);

	/*  Unsupported for now */
	default:
		break;
	}

	netdev_warn(to->dev, "Unhandled NETQUEUE OP %d\n", op);

	return VMKNETDDI_QUEUEOPS_ERR;
}

static void tg3_netq_init(struct tg3 *tp)
{
	if (!(tp->tg3_flags3 & TG3_FLG3_IOV_CAPABLE))
		return;

	VMKNETDDI_REGISTER_QUEUEOPS(tp->dev, tg3_netqueue_ops);
	netdev_info(tp->dev, "VMware NetQueue Ops is registered\n");
}

static void tg3_netq_restore(struct tg3 *tp)
{
	if (!(tp->tg3_flags3 & TG3_FLG3_IOV_CAPABLE))
		return;

	/* Enable the VRQs */
	tw32(MAC_VRQ_ENABLE, MAC_VRQ_ENABLE_DFLT_VRQ);
}
#endif /* TG3_VMWARE_NETQ_ENABLE */
